searchNodes=[{"doc":"Job pools for Elixir OTP applications, written in Elixir. Use OddJob when you need to limit concurrency of background processing in your Elixir app, like forcing backpressure on calls to databases or external APIs. OddJob is easy to use and configure, and provides functions for fire-and-forget jobs, async/await calls where the results must be returned, and job scheduling. Job queues are stored in process memory so no database is required. Installation The package can be installed by adding odd_job to your list of dependencies in mix.exs : def deps do [ { :odd_job , &quot;~&gt; 0.4.0&quot; } ] end Getting started After installation you can start processing jobs right away. OddJob automatically starts up a supervised job pool of 5 workers out of the box with no configuration required. The default name of this job pool is OddJob.Pool , and it can be sent work in the following way: OddJob . perform ( OddJob.Pool , fn -&gt; do_some_work ( ) end ) You can skip ahead for more usage , or read on for a guide to configuring your job pools. Configuration The default pool can be customized in your config if you want to change the pool size: config :odd_job , pool_size : 10 # defaults to 5 If you are processing jobs that have a high chance of failure, you may want to customize the max_restarts and max_seconds options to prevent all the workers in a pool from restarting if too many jobs are failing. These options default to the Supervisor defaults ( max_restarts: 3, max_seconds: 5 ) and can be overridden in your config: config :odd_job , pool_size : 50 , max_restarts : 10 , max_seconds : 2 Extra pools You can add extra pools to be supervised by the OddJob application supervision tree: config :odd_job , extra_pools : [ MyApp.Email , MyApp.ExternalService ] By default, extra pools will be configured with the same options as your default pool. Luckily, extra pools can receive their own list of overrides: config :odd_job , pool_size : 50 , max_restarts : 5 , extra_pools : [ MyApp.Email , # MyApp.Email will use the defaults &quot;MyApp.ExternalService&quot; : [ # the MyApp.ExternalService pool gets its own config pool_size : 10 , max_restarts : 2 ] ] Next we'll see how you can add job pools to your own application's supervision tree . If you don't want OddJob to supervise any pools for you (including the default OddJob.Pool pool) do not set a value for :extra_pools and pass false to the :default_pool config key: config :odd_job , default_pool : false Supervising job pools You can dynamically start a new job pool linked to the current process by calling OddJob.start_link/1 : { :ok , pid } = OddJob . start_link ( name : MyApp.Email , pool_size : 10 ) OddJob . perform ( MyApp.Email , fn -&gt; do_something ( ) end ) #=&gt; :ok The first argument to the function is the name of the pool, the second argument is a keyword list of options to configure the pool. See the OddJob.start_link/1 documentation for more details. In most cases you'll want to supervise your job pools, which you can do by adding a tuple in the form of {OddJob, options} directly to the top level of your application's supervision tree or any other list of child specs for a supervisor: defmodule MyApp.Application do use Application def start ( _type , _args ) do children = [ { OddJob , name : MyApp.Email } , { OddJob , name : MyApp.ExternalService } ] opts = [ strategy : :one_for_one , name : MyApp.Supervisor ] Supervisor . start_link ( children , opts ) end end The tuple {OddJob, name: MyApp.Email} will return a child spec for a supervisor that will start and supervise the MyApp.Email pool. The second element of the tuple must be a keyword list of options with a :name key and a unique name value as an atom. You can supervise as many pools as you want, as long as they have unique names. Any default configuration options listed in your config.exs will also apply to your own supervised pools. You can override the config for any pool by passing specifying the configuration in your child spec options: children = [ # The MyApp.Email pool will use the default config: { OddJob , name : MyApp.Email } , # The MyApp.ExternalService pool will not: { OddJob , name : MyApp.ExternalService , pool_size : 20 , max_restarts : 10 } ] Module-based pools You may want to configure your pool at runtime, or wrap your logic in a custom API. Module-based pools are great for this. Invoking use OddJob.Pool defines a child_spec/1 function that can be used to start your pool under a supervisor. Imagine you want to start a job pool with a dynamically configurable pool size and wrap it in a client API: defmodule MyApp.Email do use OddJob.Pool def start_link ( init_arg ) do OddJob . start_link ( name : __MODULE__ , pool_size : init_arg ) end # Client API def send_email ( user ) do OddJob . perform ( __MODULE__ , fn -&gt; MyApp.Mailer . send ( user ) end ) end end Now you can supervise your pool and set the pool size in a child spec tuple: children = [ { MyApp.Email , 20 } ] Supervisor . start_link ( children , strategy : :one_for_one ) You can also skip the initial argument by passing MyApp.Email on its own: # in my_app/application.ex children = [ MyApp.Email # Same as {MyApp.Email, []} ] Supervisor . start_link ( children , strategy : :one_for_one ) # in my_app/job.ex defmodule MyApp.Email do use OddJob.Pool def start_link ( _init_arg ) do OddJob . start_link ( __MODULE__ ) end end For convenience, use OddJob.Pool automatically defines an overridable start_link/1 function just like the one above, that ignores the initial argument and names the pool after the module, using the default configuration options. You can pass any supervision start options to use OddJob.Pool : use OddJob.Pool , restart : :transient , shutdown : :brutal_kill See the Supervisor module for more info on supervision start options. All of the previously mentioned config options can be combined. You can have a default pool, extra pools in the OddJob supervision tree, and pools to be supervised by your own application, all of which can either use the default config or their own overrides. Usage A job pool can be sent jobs by passing its unique name and an anonymous function to one of the OddJob module's perform functions: job = OddJob . async_perform ( MyApp.ExternalService , fn -&gt; get_data ( user ) end ) # do something else data = OddJob . await ( job ) OddJob . perform ( MyApp.Email , fn -&gt; send_email ( user , data ) end ) If a worker in the pool is available then the job will be performed right away. If all of the workers are already assigned to other jobs then the new job will be added to a FIFO queue. Jobs in the queue are performed as workers become available. Use perform/2 for immediate fire and forget jobs where you don't care about the results or if it succeeds. async_perform/2 and await/1 follow the async/await pattern in the Task module, and are useful when you need to retrieve the results and you care about success or failure. Similarly to Task.async/1 , async jobs will be linked and monitored by the caller (in this case, through a proxy). If either the caller or the job crash or exit, the other will crash or exit with the same reason. Scheduled jobs Jobs can be scheduled for later execution with perform_after/3 and perform_at/3 : OddJob . perform_after ( 1_000_000 , OddJob.Pool , fn -&gt; clean_database ( ) end ) # accepts a timer in milliseconds time = DateTime . utc_now |&gt; DateTime . add ( 60 * 60 * 24 , :second ) # 24 hours from now OddJob . perform_at ( time , OddJob.Pool , fn -&gt; verify_work_is_done ( ) end ) # accepts a future DateTime struct The scheduling functions return a unique timer reference which can be read with Process.read_timer/ 1 and cancelled with OddJob.cancel_timer/1 , which will cancel execution of the job itself and clean up after itself by causing the scheduler process to exit. When the timer is up the job will be sent to the pool and can no longer be aborted. ref = OddJob . perform_after ( 5000 , OddJob.Pool , fn -&gt; :will_be_canceled end ) # somewhere else in your code if some_condition ( ) do OddJob . cancel_timer ( ref ) end Note that there is no guarantee that a scheduled job will be executed immediately when the timer runs out. Like all jobs it is sent to the pool and if all workers are busy then the job enters the queue to be performed as soon as a worker is available. License MIT - Copyright (c) 2022 M. Simon Borg","ref":"OddJob.html","title":"OddJob","type":"module"},{"doc":"Performs an async job that can be awaited on for the result. Returns the OddJob.Job.t() struct if the pool server exists at the time of calling, or {:error, :not_found} if it doesn't. Functions similarly to Task.async/1 and Task.await/2 with some tweaks. Since the workers in the pool already exist when this function is called and may be busy performing other jobs, it is unknown at the time of calling when the job will be performed and by which worker process. An OddJob.Async.Proxy is dynamically created and linked and monitored by the caller. The Proxy handles the transactions with the pool and is notified when a worker is assigned. The proxy links and monitors the worker immediately before performance of the job, forwarding any exits or crashes to the caller. The worker returns the results to the proxy, which then passes them to the caller with a {ref, result} message. The proxy immediately exits with reason :normal after the results are sent back to the caller. See await/2 and await_many/2 for functions to retrieve the results, and the Task module for more information on the async/await patter and when you may want to use it. Examples iex&gt; job = OddJob . async_perform ( OddJob.Pool , fn -&gt; :math . exp ( 100 ) end ) iex&gt; OddJob . await ( job ) 2.6881171418161356e43 iex&gt; OddJob . async_perform ( :does_not_exist , fn -&gt; &quot;never called&quot; end ) { :error , :not_found }","ref":"OddJob.html#async_perform/2","title":"OddJob.async_perform/2","type":"function"},{"doc":"Sends a collection of async jobs to the pool. Returns a list of OddJob.Job.t() structs if the pool server exists at the time of calling, or {:error, :not_found} if it doesn't. Enumerates over the collection, using each member as the argument to an anonymous function with an arity of 1 . See async_perform/2 and perform/2 for more information. There's a limit to the number of jobs that can be started with this function that roughly equals the BEAM's process limit. Examples iex&gt; jobs = OddJob . async_perform_many ( OddJob.Pool , 1 .. 5 , fn x -&gt; x * * 2 end ) iex&gt; OddJob . await_many ( jobs ) [ 1 , 4 , 9 , 16 , 25 ] iex&gt; OddJob . async_perform_many ( :does_not_exist , [ &quot;never&quot; , &quot;called&quot; ] , fn x -&gt; x end ) { :error , :not_found }","ref":"OddJob.html#async_perform_many/3","title":"OddJob.async_perform_many/3","type":"function"},{"doc":"Awaits on an async job reply and returns the results. The current process and the worker process are linked during execution of the job by an OddJob.Async.Proxy . In case the worker process dies during execution, the current process will exit with the same reason as the worker. A timeout, in milliseconds or :infinity, can be given with a default value of 5000. If the timeout is exceeded, then the current process will exit. If the worker process is linked to the current process which is the case when a job is started with async, then the worker process will also exit. This function assumes the proxy's monitor is still active or the monitor's :DOWN message is in the message queue. If it has been demonitored, or the message already received, this function will wait for the duration of the timeout awaiting the message. Examples iex&gt; OddJob . async_perform ( OddJob.Pool , fn -&gt; :math . log ( 2.6881171418161356e43 ) end ) ...&gt; |&gt; OddJob . await ( ) 100.0","ref":"OddJob.html#await/2","title":"OddJob.await/2","type":"function"},{"doc":"Awaits replies from multiple async jobs and returns them in a list. This function receives a list of jobs and waits for their replies in the given time interval. It returns a list of the results, in the same order as the jobs supplied in the jobs input argument. The current process and each worker process are linked during execution of the job by an OddJob.Async.Proxy . If any of the worker processes dies, the caller process will exit with the same reason as that worker. A timeout, in milliseconds or :infinity , can be given with a default value of 5000 . If the timeout is exceeded, then the caller process will exit. Any worker processes that are linked to the caller process (which is the case when a job is started with async_perform/2 ) will also exit. This function assumes the proxies' monitors are still active or the monitor's :DOWN message is in the message queue. If any jobs have been demonitored, or the message already received, this function will wait for the duration of the timeout. Examples iex&gt; job1 = OddJob . async_perform ( OddJob.Pool , fn -&gt; 2 * * 2 end ) iex&gt; job2 = OddJob . async_perform ( OddJob.Pool , fn -&gt; 3 * * 2 end ) iex&gt; [ job1 , job2 ] |&gt; OddJob . await_many ( ) [ 4 , 9 ]","ref":"OddJob.html#await_many/2","title":"OddJob.await_many/2","type":"function"},{"doc":"Cancels a scheduled job. timer_ref is the unique reference returned by perform_at/3 or perform_after/3 . This function returns the number of milliseconds left in the timer when cancelled, or false if the timer already expired. If the return is false you can assume that the job has already been sent to the pool for execution. NOTE: Cancelling the timer with this function ensures that the job is never executed and that the scheduler process is exited and not left &quot;hanging&quot;. Using Process.cancel_timer/1 will also cancel execution, but may leave hanging processes. A hanging scheduler process will eventually timeout, but not until one second after the expiration of the original timer. Examples iex&gt; ref = OddJob . perform_after ( 500 , OddJob.Pool , fn -&gt; :never end ) iex&gt; time = OddJob . cancel_timer ( ref ) iex&gt; is_integer ( time ) true iex&gt; ref = OddJob . perform_after ( 10 , OddJob.Pool , fn -&gt; :never end ) iex&gt; Process . sleep ( 11 ) iex&gt; OddJob . cancel_timer ( ref ) false","ref":"OddJob.html#cancel_timer/1","title":"OddJob.cancel_timer/1","type":"function"},{"doc":"Returns a specification to start this module under a supervisor. Examples iex&gt; children = [ OddJob . child_spec ( name : MyApp.Media ) ] [ %{ id : MyApp.Media , start : { OddJob.Pool , :start_link , [ [ name : MyApp.Media ] ] } , type : :supervisor } ] iex&gt; { :ok , _pid } = Supervisor . start_link ( children , strategy : :one_for_one ) iex&gt; OddJob . workers ( MyApp.Media ) |&gt; length ( ) 5 Normally you would start an OddJob pool under a supervision tree with a child specification tuple and not call child_spec/1 directly. children = [ { OddJob , name : MyApp.Media } ] Supervisor . start_link ( children , strategy : :one_for_one ) You must pass a keyword list of options to child_spec/1 , or as the second element of the child specification tuple. Options :name - An atom that names the pool. This argument is required. :pool_size - A positive integer , the number of concurrent workers in the pool. Defaults to 5 or your application's config value. :max_restarts - A positive integer , the number of worker restarts allowed in a given timeframe before all of the workers are restarted. Set a higher number if your jobs have a high rate of expected failure. Defaults to 5 or your application's config value. :max_seconds - A positive integer , the timeframe in seconds in which max_restarts applies. Defaults to 3 or your application's config value. See Supervisor for more info on restart intensity options. See the Supervisor module for more about child specs.","ref":"OddJob.html#child_spec/1","title":"OddJob.child_spec/1","type":"function"},{"doc":"Performs a fire and forget job. Casts the function to the job pool , and if a worker is available it is performed immediately. If no workers are available the job is added to the back of a FIFO queue to be performed as soon as a worker is available. Returns :ok if the pool process exists at the time of calling, or {:error, :not_found} if it doesn't. The job is casted to the pool, so there can be no guarantee that the pool process will still exist if it crashes between the beginning of the function call and when the message is received. Use this function for interaction with other processes or the outside world where you don't care about the return results or whether it succeeds. Examples iex&gt; parent = self ( ) iex&gt; :ok = OddJob . perform ( OddJob.Pool , fn -&gt; send ( parent , :hello ) end ) iex&gt; receive do ...&gt; msg -&gt; msg ...&gt; end :hello iex&gt; OddJob . perform ( :does_not_exist , fn -&gt; &quot;never called&quot; end ) { :error , :not_found }","ref":"OddJob.html#perform/2","title":"OddJob.perform/2","type":"function"},{"doc":"Sends a job to the pool after the given timer has elapsed. timer is an integer that indicates the number of milliseconds that should elapse before the job is sent to the pool. The timed message is executed under a separate supervised process, so if the current process crashes the job will still be performed. A timer reference is returned, which can be read with Process.read_timer/1 or cancelled with OddJob.cancel_timer/1 . Returns {:error, :not_found} if the pool server does not exist at the time this function is called. Examples timer_ref = OddJob . perform_after ( 5000 , OddJob.Pool , fn -&gt; deferred_job ( ) end ) Process . read_timer ( timer_ref ) #=&gt; 2836 # time remaining before job is sent to the pool OddJob . cancel_timer ( timer_ref ) #=&gt; 1175 # job has been cancelled timer_ref = OddJob . perform_after ( 5000 , OddJob.Pool , fn -&gt; deferred_job ( ) end ) Process . sleep ( 6000 ) OddJob . cancel_timer ( timer_ref ) #=&gt; false # too much time has passed to cancel the job iex&gt; OddJob . perform_after ( 5000 , :does_not_exist , fn -&gt; &quot;never called&quot; end ) { :error , :not_found }","ref":"OddJob.html#perform_after/3","title":"OddJob.perform_after/3","type":"function"},{"doc":"Sends a job to the pool at the given time . time must be a DateTime struct for a time in the future. The timer is executed under a separate supervised process, so if the caller crashes the job will still be performed. A timer reference is returned, which can be read with Process.read_timer/1 or canceled with OddJob.cancel_timer/1 . Returns {:error, :not_found} if the pool server does not exist at the time this function is called. Examples time = DateTime . utc_now ( ) |&gt; DateTime . add ( 600 , :second ) OddJob . perform_at ( time , OddJob.Pool , fn -&gt; scheduled_job ( ) end ) iex&gt; time = DateTime . utc_now ( ) |&gt; DateTime . add ( 600 , :second ) iex&gt; OddJob . perform_at ( time , :does_not_exist , fn -&gt; &quot;never called&quot; end ) { :error , :not_found }","ref":"OddJob.html#perform_at/3","title":"OddJob.perform_at/3","type":"function"},{"doc":"Sends a collection of jobs to the pool. Enumerates over the collection, using each member as the argument to an anonymous function witn an arity of 1 . The collection of jobs is sent in bulk to the pool where they will be immediately performed by any available workers. If there are no available workers the jobs are added to the FIFO queue. Jobs are guaranteed to always process in the same order of the original collection, and insertion into the queue is atomic. No jobs sent from other processes can be inserted between any member of the collection. Returns :ok if the pool process exists at the time of calling, or {:error, :not_found} if it doesn't. See perform/2 for more. Examples iex&gt; caller = self ( ) iex&gt; OddJob . perform_many ( OddJob.Pool , 1 .. 5 , fn x -&gt; send ( caller , x ) end ) iex&gt; for x &lt;- 1 .. 5 do ...&gt; receive do ...&gt; ^ x -&gt; x ...&gt; end ...&gt; end [ 1 , 2 , 3 , 4 , 5 ] iex&gt; OddJob . perform_many ( :does_not_exist , [ &quot;never&quot; , &quot;called&quot; ] , fn x -&gt; x end ) { :error , :not_found }","ref":"OddJob.html#perform_many/3","title":"OddJob.perform_many/3","type":"function"},{"doc":"Sends a collection of jobs to the pool after the given timer has elapsed. Enumerates over the collection, using each member as the argument to an anonymous function with an arity of 1 . Returns a single timer reference . The timer is watched by a single process that will send the entire batch of jobs to the pool when the timer expires. See perform_after/3 for more information about arguments and timers. Consider using this function instead of perform_after/3 when scheduling a large batch of jobs. perform_after/3 starts a separate scheduler process per job, whereas this function starts a single scheduler process for the whole batch. Returns {:error, :not_found} if the pool does not exist at the time the function is called. Examples timer_ref = OddJob . perform_many_after ( 5000 , OddJob.Pool , 1 .. 5 , fn x -&gt; x * * 2 end ) #=&gt; #Reference&lt;0.1431903625.286261254.39156&gt; # fewer than 5 seconds pass OddJob . cancel_timer ( timer_ref ) #=&gt; 2554 # returns the time remaining. All jobs in the collection have been canceled timer_ref = OddJob . perform_many_after ( 5000 , OddJob.Pool , 1 .. 5 , fn x -&gt; x * * 2 end ) #=&gt; #Reference&lt;0.1431903625.286261254.39178&gt; # more than 5 seconds pass OddJob . cancel_timer ( timer_ref ) #=&gt; false # too much time has passed, all jobs have been queued iex&gt; OddJob . perform_many_after ( 5000 , :does_not_exist , [ &quot;never&quot; , &quot;called&quot; ] , fn x -&gt; x end ) { :error , :not_found }","ref":"OddJob.html#perform_many_after/4","title":"OddJob.perform_many_after/4","type":"function"},{"doc":"Sends a collection of jobs to the pool at the given time . Enumerates over the collection, using each member as the argument to an anonymous function with an arity of 1 . Returns a single timer reference . The timer is watched by a single process that will send the entire batch of jobs to the pool when the timer expires. See perform_at/3 for more information about arguments and timers. Consider using this function instead of perform_at/3 when scheduling a large batch of jobs. perform_at/3 starts a separate scheduler process per job, whereas this function starts a single scheduler process for the whole batch. Returns {:error, :not_found} if the pool does not exist at the time the function is called. Examples time = DateTime . utc_now ( ) |&gt; DateTime . add ( 5 , :second ) timer_ref = OddJob . perform_many_at ( time , OddJob.Pool , 1 .. 5 , fn x -&gt; x * * 2 end ) #=&gt; #Reference&lt;0.1431903625.286261254.39156&gt; # fewer than 5 seconds pass OddJob . cancel_timer ( timer_ref ) #=&gt; 2554 # returns the time remaining. All jobs in the collection have been canceled time = DateTime . utc_now ( ) |&gt; DateTime . add ( 5 , :second ) timer_ref = OddJob . perform_many_at ( time , OddJob.Pool , 1 .. 5 , fn x -&gt; x * * 2 end ) #=&gt; #Reference&lt;0.1431903625.286261254.39178&gt; # more than 5 seconds pass OddJob . cancel_timer ( timer_ref ) #=&gt; false # too much time has passed, all jobs have been queued iex&gt; time = DateTime . utc_now ( ) |&gt; DateTime . add ( 5 , :second ) iex&gt; OddJob . perform_many_at ( time , :does_not_exist , [ &quot;never&quot; , &quot;called&quot; ] , fn x -&gt; x end ) { :error , :not_found }","ref":"OddJob.html#perform_many_at/4","title":"OddJob.perform_many_at/4","type":"function"},{"doc":"A macro for creating jobs with an expressive DSL. perform_this/2 works like perform/2 except it accepts a do block instead of an anonymous function. Examples You must import or require OddJob to use macros: import OddJob alias MyApp.Job perform_this Job do some_work ( ) some_other_work ( ) end perform_this Job , do : something_hard ( )","ref":"OddJob.html#perform_this/2","title":"OddJob.perform_this/2","type":"macro"},{"doc":"A macro for creating jobs with an expressive DSL. perform_this/3 accepts a single configuration option as the second argument that will control execution of the job. The available options provide the functionality of async_perform/2 , perform_at/3 , and perform_after/3 . Options :async - Passing the atom :async as the second argument before the do block creates an async job that can be awaited on. See async_perform/2 . at: time - Use this option to schedule the job for a specific time in the future. time must be a valid Time or DateTime struct. See perform_at/3 . after: timer - Use this option to schedule the job to perform after the given timer has elapsed. timer must be in milliseconds. See perform_after/3 . Examples import OddJob alias MyApp.Job time = ~T[03:00:00.000000] perform_this Job , at : time do scheduled_work ( ) end perform_this Job , after : 5000 , do : something_important ( ) perform_this Job , :async do get_data ( ) end |&gt; await ( ) iex&gt; ( perform_this OddJob.Pool , :async , do : 10 * * 2 ) |&gt; await ( ) 100","ref":"OddJob.html#perform_this/3","title":"OddJob.perform_this/3","type":"macro"},{"doc":"Returns the pid of the OddJob.Pool.Supervisor that supervises the job pool 's workers. Returns {:error, :not_found} if the process does not exist at the time this function is called. There is no guarantee that the process identified by the pid will still be alive after the results are returned, as it could exit or be killed or restarted at any time. Use pool_supervisor_name/1 to obtain the persistent registered name of the supervisor. Examples OddJob . pool_supervisor ( OddJob.Pool ) #=&gt; #PID&lt;0.239.0&gt; iex&gt; OddJob . pool_supervisor ( :does_not_exist ) { :error , :not_found }","ref":"OddJob.html#pool_supervisor/1","title":"OddJob.pool_supervisor/1","type":"function"},{"doc":"Returns the registered name in :via of the OddJob.Pool.Supervisor that supervises the pool 's workers Returns {:error, :not_found} if the process does not exist at the time this function is called. Examples iex&gt; OddJob . pool_supervisor_name ( OddJob.Pool ) { :via , Registry , { OddJob.Registry , { OddJob.Pool , :pool_sup } } } iex&gt; OddJob . pool_supervisor_name ( :does_not_exist ) { :error , :not_found }","ref":"OddJob.html#pool_supervisor_name/1","title":"OddJob.pool_supervisor_name/1","type":"function"},{"doc":"Returns a two element tuple {pid, state} with the pid of the job queue as the first element. The second element is the state of the queue represented by an OddJob.Queue struct. The queue state has the following fields: :workers - A list of the pid s of all workers in the pool :assigned - The pid s of the workers currently assigned to jobs :jobs - A list of OddJob.Job structs in the queue awaiting execution :pool - The Elixir term naming the pool that the queue belongs to Returns {:error, :not_found} if the queue server does not exist at the time this function is called. Examples iex&gt; { pid , % OddJob.Queue { pool : pool } } = OddJob . queue ( OddJob.Pool ) iex&gt; is_pid ( pid ) true iex&gt; pool OddJob.Pool iex&gt; OddJob . queue ( :does_not_exist ) { :error , :not_found }","ref":"OddJob.html#queue/1","title":"OddJob.queue/1","type":"function"},{"doc":"Returns the registered name in :via for the queue process. See the Registry module for more information on :via process name registration. Returns {:error, :not_found} if the queue server does not exist at the time this function is called. Examples iex&gt; OddJob . queue_name ( OddJob.Pool ) { :via , Registry , { OddJob.Registry , { OddJob.Pool , :queue } } } iex&gt; OddJob . queue_name ( :does_not_exist ) { :error , :not_found }","ref":"OddJob.html#queue_name/1","title":"OddJob.queue_name/1","type":"function"},{"doc":"Starts an OddJob.Pool supervision tree linked to the current process. You can start an OddJob pool dynamically with start_link/1 to start processing concurrent jobs: iex&gt; { :ok , _pid } = OddJob . start_link ( name : MyApp.Event , pool_size : 10 ) iex&gt; OddJob . async_perform ( MyApp.Event , fn -&gt; :do_something end ) |&gt; OddJob . await ( ) :do_something In most cases you would instead start your pools inside a supervision tree: children = [ { OddJob , name : MyApp.Event , pool_size : 10 } ] Supervisor . start_link ( children , strategy : :one_for_one ) See OddJob.child_spec/1 for a list of available options. See Supervisor for more on child specifications and supervision trees.","ref":"OddJob.html#start_link/1","title":"OddJob.start_link/1","type":"function"},{"doc":"Returns the pid of the top level supervisor in the pool supervision tree, nil if a process can't be found. Example iex&gt; { :ok , pid } = OddJob . start_link ( name : :whereis ) iex&gt; OddJob . whereis ( :whereis ) == pid true iex&gt; OddJob . whereis ( :where_it_is_not ) == nil true","ref":"OddJob.html#whereis/1","title":"OddJob.whereis/1","type":"function"},{"doc":"Returns a list of pid s for the specified worker pool. There is no guarantee that the processes will still be alive after the results are returned, as they could exit or be killed at any time. Returns {:error, :not_found} if the pool does not exist at the time this function is called. Examples OddJob . workers ( OddJob.Pool ) #=&gt; [#PID&lt;0.105.0&gt;, #PID&lt;0.106.0&gt;, #PID&lt;0.107.0&gt;, #PID&lt;0.108.0&gt;, #PID&lt;0.109.0&gt;] iex&gt; OddJob . workers ( :does_not_exist ) { :error , :not_found }","ref":"OddJob.html#workers/1","title":"OddJob.workers/1","type":"function"},{"doc":"","ref":"OddJob.html#t:child_spec/0","title":"OddJob.child_spec/0","type":"type"},{"doc":"","ref":"OddJob.html#t:date_time/0","title":"OddJob.date_time/0","type":"type"},{"doc":"","ref":"OddJob.html#t:job/0","title":"OddJob.job/0","type":"type"},{"doc":"","ref":"OddJob.html#t:name/0","title":"OddJob.name/0","type":"type"},{"doc":"","ref":"OddJob.html#t:not_found/0","title":"OddJob.not_found/0","type":"type"},{"doc":"","ref":"OddJob.html#t:options/0","title":"OddJob.options/0","type":"type"},{"doc":"","ref":"OddJob.html#t:pool/0","title":"OddJob.pool/0","type":"type"},{"doc":"","ref":"OddJob.html#t:queue/0","title":"OddJob.queue/0","type":"type"},{"doc":"","ref":"OddJob.html#t:start_option/0","title":"OddJob.start_option/0","type":"type"},{"doc":"","ref":"OddJob.html#t:timer/0","title":"OddJob.timer/0","type":"type"},{"doc":"Internal utilities for working with OddJob processes. To avoid unexpected behavior, it's recommended that developers do not interact directly with OddJob processes. See OddJob for the public API.","ref":"OddJob.Utils.html","title":"OddJob.Utils","type":"module"},{"doc":"Extracts the pid from an {:ok, pid} tuple. iex&gt; pid = spawn_link ( fn -&gt; :hello end ) iex&gt; pid == OddJob.Utils . extract_pid ( { :ok , pid } ) true","ref":"OddJob.Utils.html#extract_pid/1","title":"OddJob.Utils.extract_pid/1","type":"function"},{"doc":"Links and monitors the pid and returns the tuple {pid, monitor_ref} . Example iex&gt; pid = spawn ( fn -&gt; [ ] end ) iex&gt; { ^ pid , ref } = OddJob.Utils . link_and_monitor ( pid ) iex&gt; receive do ...&gt; { :DOWN , ^ ref , :process , ^ pid , :normal } -&gt; :received ...&gt; end :received","ref":"OddJob.Utils.html#link_and_monitor/1","title":"OddJob.Utils.link_and_monitor/1","type":"function"},{"doc":"Returns the OddJob.Pool.Supervisor name in :via for the given name . iex&gt; OddJob.Utils . pool_supervisor_name ( :job ) { :via , Registry , { OddJob.Registry , { :job , :pool_sup } } }","ref":"OddJob.Utils.html#pool_supervisor_name/1","title":"OddJob.Utils.pool_supervisor_name/1","type":"function"},{"doc":"Returns the OddJob.Async.ProxySupervisor name in :via for the given name . iex&gt; OddJob.Utils . proxy_sup_name ( :job ) { :via , Registry , { OddJob.Registry , { :job , :proxy_sup } } }","ref":"OddJob.Utils.html#proxy_sup_name/1","title":"OddJob.Utils.proxy_sup_name/1","type":"function"},{"doc":"Returns the OddJob.Queue name in :via for the given name . iex&gt; OddJob.Utils . queue_name ( :job ) { :via , Registry , { OddJob.Registry , { :job , :queue } } }","ref":"OddJob.Utils.html#queue_name/1","title":"OddJob.Utils.queue_name/1","type":"function"},{"doc":"Returns the OddJob.Scheduler.Supervisor name in :via for the given name . iex&gt; OddJob.Utils . scheduler_sup_name ( :job ) { :via , Registry , { OddJob.Registry , { :job , :scheduler_sup } } }","ref":"OddJob.Utils.html#scheduler_sup_name/1","title":"OddJob.Utils.scheduler_sup_name/1","type":"function"},{"doc":"","ref":"OddJob.Utils.html#t:name/0","title":"OddJob.Utils.name/0","type":"type"},{"doc":"The OddJob.Job struct holds all of the useful information about a job.","ref":"OddJob.Job.html","title":"OddJob.Job","type":"module"},{"doc":"The OddJob.Job struct is the datatype that is passed between processes charged with performing the job. It holds all of the data that is necessary to link, monitor, perform work, and return results to the caller. The job struct is only returned to the caller when using the async/await pattern. When the caller receives the struct after calling OddJob.async_perform/2 the :results field is always nil , even though the work could conceivably already be done. This is because the results are not waited on at the time the struct is created. The results are only known when passing the job to OddJob.await/2 or matching on the {ref, results} message. :function is the anonymous function that will be performed by the worker :results is the term that is returned by function . This is only used internally by the processes performing the work. :async is an boolean identifying if the job's results can be awaited on :ref is the unique monitor reference of the job :owner is the pid of the calling process, i.e. self() :proxy is the pid of the proxy server that creates the job and routes the results. The owner links and monitors the proxy , while the proxy links and monitors the worker. Exit messages and failures cascade up to the owner . The worker sends results back to the proxy , which then sends them to the owner before exiting with reason :normal .","ref":"OddJob.Job.html#t:t/0","title":"OddJob.Job.t/0","type":"type"},{"doc":"The OddJob.Async.Proxy links the job caller to the worker as the job is being performed. The process that calls async_perform/2 or async_perform_many/3 must link and monitor the worker performing the job so it can receive the results and be notified of failure by exiting or receiving an :EXIT message if the process is trapping exits. However, at the time the function is called it is unknown which worker will receive the job and when. The proxy is a process that is created to provide a link between the caller and the worker. When an async funcion is called, the caller links and monitors the proxy, and the proxy pid and monitor reference are stored in the OddJob.Job struct. When the worker receives the job it calls the proxy to be linked and monitored. Job results as well as :EXIT and :DOWN messages are passed from the worker to the caller via the proxy.","ref":"OddJob.Async.Proxy.html","title":"OddJob.Async.Proxy","type":"module"},{"doc":"","ref":"OddJob.Async.Proxy.html#t:job/0","title":"OddJob.Async.Proxy.job/0","type":"type"},{"doc":"","ref":"OddJob.Async.Proxy.html#t:t/0","title":"OddJob.Async.Proxy.t/0","type":"type"},{"doc":"The OddJob.Async.ProxySupervisor is a DynamicSupervisor that starts and stops children OddJob.Async.Proxy processes.","ref":"OddJob.Async.ProxySupervisor.html","title":"OddJob.Async.ProxySupervisor","type":"module"},{"doc":"","ref":"OddJob.Async.ProxySupervisor.html#start_child/1","title":"OddJob.Async.ProxySupervisor.start_child/1","type":"function"},{"doc":"The OddJob.Pool process The OddJob.Pool is the supervisor at the top of a pool supervision tree, and is responsible for starting and supervising the OddJob.Pool.Supervisor , OddJob.Queue , OddJob.Async.ProxySupervisor , and OddJob.Scheduler.Supervisor . When you add a OddJob.child_spec/1 to your list of supervised children, or call OddJob.start_link/1 , you are adding an OddJob.Pool to your application. Public functions This module defines the public functions start_link/1 and child_spec/1 . It is recommended that you call these functions from the OddJob module namespace instead. See OddJob for documentation and usage. Module-based pools You may want to configure your pool at runtime, or wrap your logic in a custom API. Module-based pools are great for this. Invoking use OddJob.Pool defines a child_spec/1 function that can be used to start your pool under a supervisor. Imagine you want to start a job pool with a dynamically configurable pool size and wrap it in a client API: defmodule MyApp.Email do use OddJob.Pool def start_link ( init_arg ) do OddJob . start_link ( name : __MODULE__ , pool_size : init_arg ) end # Client API def send_email ( user ) do OddJob . perform ( __MODULE__ , fn -&gt; MyApp.Mailer . send ( user ) end ) end end Now you can supervise your pool and set the pool size in a child spec tuple: children = [ { MyApp.Email , 20 } ] Supervisor . start_link ( children , strategy : :one_for_one ) You can also skip the initial argument by passing MyApp.Email on its own: # in my_app/application.ex children = [ MyApp.Email # Same as {MyApp.Email, []} ] Supervisor . start_link ( children , strategy : :one_for_one ) # in my_app/job.ex defmodule MyApp.Email do use OddJob.Pool def start_link ( _init_arg ) do OddJob . start_link ( __MODULE__ ) end end For convenience, use OddJob.Pool automatically defines an overridable start_link/1 function just like the one above, that ignores the initial argument and names the pool after the module, using the default configuration options. You can pass any supervision start options to use OddJob.Pool : use OddJob.Pool , restart : :transient , shutdown : :brutal_kill See the Supervisor module for more info on supervision start options.","ref":"OddJob.Pool.html","title":"OddJob.Pool","type":"module"},{"doc":"","ref":"OddJob.Pool.html#t:child_spec/0","title":"OddJob.Pool.child_spec/0","type":"type"},{"doc":"","ref":"OddJob.Pool.html#t:name/0","title":"OddJob.Pool.name/0","type":"type"},{"doc":"","ref":"OddJob.Pool.html#t:options/0","title":"OddJob.Pool.options/0","type":"type"},{"doc":"","ref":"OddJob.Pool.html#t:start_option/0","title":"OddJob.Pool.start_option/0","type":"type"},{"doc":"The OddJob.Pool.Supervisor supervises the pool of OddJob.Pool.Worker s.","ref":"OddJob.Pool.Supervisor.html","title":"OddJob.Pool.Supervisor","type":"module"},{"doc":"The OddJob.Pool.Worker is a GenServer that performs concurrent work as one of many members of an OddJob.Pool . The OddJob.Pool.Worker checks in with the pool and asks to be monitored upon startup. Once the worker is monitored it can start receiving jobs. In the unlikely event that the pool crashes, the worker will be notified and request to be monitored again when the pool restarts.","ref":"OddJob.Pool.Worker.html","title":"OddJob.Pool.Worker","type":"module"},{"doc":"","ref":"OddJob.Pool.Worker.html#t:t/0","title":"OddJob.Pool.Worker.t/0","type":"type"},{"doc":"The OddJob.Queue is a GenServer that manages the assignments given to the pool workers. The queue receives jobs and assigns them to available workers. If all workers in a pool are currently busy then new jobs are added to a FIFO queue to be processed as workers become available.","ref":"OddJob.Queue.html","title":"OddJob.Queue","type":"module"},{"doc":"","ref":"OddJob.Queue.html#t:job/0","title":"OddJob.Queue.job/0","type":"type"},{"doc":"The OddJob.Queue struct holds the state of the job queue. :pool is a term representing the name of the job pool that the queue belongs to :workers is a list of the active worker pid s, whether they are busy working or not :assigned is a list of the worker pid s that are currently assigned to a job :jobs is a list of OddJob.Job structs representing the jobs that are queued to be performed when workers are available","ref":"OddJob.Queue.html#t:t/0","title":"OddJob.Queue.t/0","type":"type"},{"doc":"The OddJob process registry.","ref":"OddJob.Registry.html","title":"OddJob.Registry","type":"module"},{"doc":"Returns the name in :via that can be used to lookup an OddJob process. The first argument is the atom that was used to name the pool . The second argument is the role as an atom, or the tuple {:worker, id} . Roles :queue - the pool process, a GenServer that receives jobs and assigns them to the workers :pool_sup - the supervisor responsible for starting and stopping the workers in the pool :proxy_sup - the supervisor responsible for starting the proxy processes that link the calling process to the worker during execution of an async job :scheduler_sup - the supervisor responsible for starting job scheduling processes Example iex&gt; OddJob.Registry . via ( ViaTest , :sup ) { :via , Registry , { OddJob.Registry , { ViaTest , :sup } } } See OddJob.Utils for helper functions that encapsulate each role.","ref":"OddJob.Registry.html#via/2","title":"OddJob.Registry.via/2","type":"function"},{"doc":"","ref":"OddJob.Registry.html#t:name/0","title":"OddJob.Registry.name/0","type":"type"},{"doc":"","ref":"OddJob.Registry.html#t:pool/0","title":"OddJob.Registry.pool/0","type":"type"},{"doc":"","ref":"OddJob.Registry.html#t:role/0","title":"OddJob.Registry.role/0","type":"type"},{"doc":"The OddJob.Scheduler is responsible for execution of scheduled jobs. Each scheduler is a dynamically supervised process that is created to manage a single timer and a job or collection of jobs to send to a pool when the timer expires. After the jobs are delivered to the pool the scheduler shuts itself down. The scheduler process will also automatically shutdown if a timer is cancelled with OddJob.cancel_timer/1 . If a timer is cancelled with Process.cancel_timer/1 then the scheduler will eventually timeout and shutdown one second after the timer would have expired.","ref":"OddJob.Scheduler.html","title":"OddJob.Scheduler","type":"module"},{"doc":"","ref":"OddJob.Scheduler.html#perform_many/4","title":"OddJob.Scheduler.perform_many/4","type":"function"},{"doc":"","ref":"OddJob.Scheduler.html#t:pool/0","title":"OddJob.Scheduler.pool/0","type":"type"},{"doc":"The OddJob.Scheduler.Supervisor is a DynamicSupervisor that starts and stops children OddJob.Scheduler processes.","ref":"OddJob.Scheduler.Supervisor.html","title":"OddJob.Scheduler.Supervisor","type":"module"},{"doc":"","ref":"OddJob.Scheduler.Supervisor.html#start_child/1","title":"OddJob.Scheduler.Supervisor.start_child/1","type":"function"},{"doc":"","ref":"changelog.html","title":"Changelog","type":"extras"},{"doc":"Improvements Hibernate OddJob.Queue after a 10 second timeout to trigger garbage collection and reduce memory consumption Fix typo in CHANGELOG.md","ref":"changelog.html#v-0-4-1","title":"Changelog - v.0.4.1","type":"extras"},{"doc":"Changes OddJob.child_spec/1 accepts a keyword list of options that must contain a :name key with an atom value, plus optional overrides. Change OddJob.supervisor_id to OddJob.pool_supervisor_name/1 Change OddJob.supervisor to OddJob.pool_supervisor/1 Change OddJob.pool_id to OddJob.queue_name/1 Change OddJob.Async.ProxyServer to OddJob.Async.Proxy In the OddJob module perform/2 , perform_many/3 , async_perform/2 , async_perform_many/3 , perform_after/3 , perform_many_after/4 , perform_at/3 , perform_many_at/4 , queue/1 , queue_name/1 , pool_supervisor/1 , pool_supervisor_name/1 , and workers/1 will now return {:error, :not_found} if the pool does not exist at the time of calling. OddJob.perform_at/3 now only accepts a DateTime struct as the first argument. Revert previous pool usage in functions to queue , and change OddJob.Pool to OddJob.Queue Default pool cannot be renamed. Additions New functions perform_many/3 , async_perform_many/3 , perform_many_after/4 , and perform_many_at/4 can more efficiently enqueue large batches of jobs New public function OddJob.start_link/1 can be used to start OddJob pools use OddJob.Pool to create module based job pools with runtime configuration and custom start options for the top level supervisor New options for max_restarts and max_seconds join pool_size as configuration for the worker supervisor Any pool can override the default config and have its own configuration Config can be set for user-supervised pools by passing an options list to the OddJob child spec tuple: {OddJob, name: :work, pool_size: 10} , or to OddJob.start_link/1 for module-based pools. Improvements Use :via process naming for dynamically named processes. Every pool supervises its own proxy supervisor and scheduler supervisor to increase isolation between pools. Better documentation about version history. Better module documentation.","ref":"changelog.html#v0-4-0","title":"Changelog - v0.4.0","type":"extras"},{"doc":"Fix a typo in documentation.","ref":"changelog.html#v0-3-3","title":"Changelog - v0.3.3","type":"extras"},{"doc":"Improve documentation of perform_this macros.","ref":"changelog.html#v0-3-2","title":"Changelog - v0.3.2","type":"extras"},{"doc":"Fix a mistake in documentation","ref":"changelog.html#v0-3-1","title":"Changelog - v0.3.1","type":"extras"},{"doc":"Remove __using__ macro from OddJob module Change to_perform_this macro name to perform_this Add :at and :after keyword options to perform_this for scheduling jobs Fix bug that causes pool to overmonitor workers in the event of worker supervisor shutdown Isolate every perform_after / perform_at schedule in its own process so a scheduling failure doesn't destroy all schedules Change default pool name to :job Change config options Rename queue to pool in public functions and internal implementation Add another level of supervision for workers, so pool does not restart when workers are failing, which would cause the job queue to be lost. Improvements and fixes to documentation and tests","ref":"changelog.html#v0-3-0","title":"Changelog - v0.3.0","type":"extras"}]